from typing import List, Union

from pikachu.general import read_smiles
from pikachu.chem.structure import Structure

from raichu.domain.domain import Domain, TailoringDomain, RecognitionDomain, \
    SynthesisDomain, CarrierDomain, TerminationDomain
from raichu.central_chain_detection.label_central_chain import label_pk_central_chain, label_nrp_central_chain
from raichu.attach_to_domain import attach_to_domain_pk, attach_to_domain_nrp
from enum import Enum, unique
from pikachu.drawing.drawing import Drawer
from raichu.drawing.drawer import RaichuDrawer

@unique
class ModuleType(Enum):
    NRPS = 1
    PKS = 2

    @staticmethod
    def from_string(label: str) -> "ModuleType":
        for value in ModuleType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown module type: {label}")


@unique
class PKSModuleSubtype(Enum):
    PKS_CIS = 1
    PKS_TRANS = 2
    PKS_ITER = 3

    @staticmethod
    def from_string(label: str) -> "PKSModuleSubtype":
        for value in PKSModuleSubtype:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown PKS module subtype: {label}")


@unique
class NRPSDomainType(Enum):
    A = 1
    C = 2
    PCP = 3
    E = 4
    nMT = 5
    TE = 6
    TD = 7
    UNKNOWN = 8

    @staticmethod
    def from_string(label: str) -> "NRPSDomainType":
        for value in NRPSDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown NRPS domain type: {label}")


@unique
class PKSDomainType(Enum):
    AT = 1
    KS = 2
    ACP = 3
    KR = 4
    DH = 5
    ER = 6
    TE = 7
    TD = 8
    UNKNOWN = 9
    DUMMY_ALMT = 10  # Alpha-L-Methyltransferase
    DUMMY_AMT = 11  # Alpha-Methyltransferase
    DUMMY_SC = 12  # Smalles cyclase for creating pyran/furan rings
    DUMMY_ZDH = 13  # E-configured double bonds
    DUMMY_EDH = 14  # Z-configured double bonds
    DUMMY_AH = 15  # Alpha-hydroxylase
    DUMMY_GDH = 16  # Gamma-beta-dehydrogenase
    DUMMY_ZGDH = 17  # Z-Gamma-beta-dehydrogenase
    DUMMY_EGDH = 18  # E-Gamma-beta-dehydrogenase
    DUMMY_OMT = 19  # Beta-Hydroxymethyltransferase
    DUMMY_BMT = 20  # Beta-Methyltransferase

    @staticmethod
    def from_string(label: str) -> "PKSDomainType":
        for value in PKSDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown PKS domain type: {label}")


class _Module:

    def __init__(self, nr: int, module_type: str, domains: List[Domain], *,
                 module_subtype: Union[str, None] = None,
                 starter: bool = False, terminator: bool = False) -> None:
        self.id = nr
        self.type = ModuleType.from_string(module_type)
        self.subtype = None

        if module_subtype is not None:
            if self.type.name == 'PKS':
                self.subtype = PKSModuleSubtype.from_string(module_subtype)
            else:
                raise ValueError(f"Module subtypes not supported for {self.type.name} modules.")

        self.domains = domains

        self.recognition_domain = None
        self.synthesis_domain = None
        self.carrier_domain = None
        self.termination_domain = None

        self.tailoring_domains = []

        self.is_starter_module = starter
        self.is_termination_module = terminator

        for domain in self.domains:
            if self.type.name == 'NRPS':
                assert domain.type.name in [v.name for v in NRPSDomainType]
            if self.type.name == 'PKS':
                assert domain.type.name in [v.name for v in PKSDomainType]

            if domain.used and domain.active:
                if isinstance(domain, TailoringDomain):
                    if domain.type.name not in [d.type.name for d in self.tailoring_domains]:
                        self.tailoring_domains.append(domain)
                    else:
                        raise ValueError(f"Cannot have two used tailoring domains of type {domain.type.name} in one \
module. Remove domain or set the 'used' or 'active' flag to False")

                if isinstance(domain, RecognitionDomain):
                    if not self.recognition_domain:
                        self.recognition_domain = domain
                    else:
                        raise ValueError(f"Cannot have more than one used recognition domain in one \
module. Remove a domain or set the 'used' or 'active' flag to False")

                if isinstance(domain, SynthesisDomain) and domain.is_elongating:
                    if not self.synthesis_domain:
                        self.synthesis_domain = domain
                    else:
                        raise ValueError("Cannot have more than one used and elongating synthesis domains \
in one module. Remove a domain, set the 'used' or 'active' flag to False, or set the 'is_elongating' flag to False.")

                if isinstance(domain, CarrierDomain):
                    if not self.carrier_domain:
                        self.carrier_domain = domain
                    else:
                        raise ValueError("Cannot have more than one used carrier domain in one \
module. Remove a domain or set the 'used' or 'active' flag to False")

                if isinstance(domain, TerminationDomain):
                    if not self.termination_domain:
                        self.termination_domain = domain
                        if not self.is_termination_module:
                            self.termination_domain.used = False
                    else:
                        raise ValueError("Cannot have more than one used termination domain in one \
module. Remove a domain or set the 'used' or 'active' flag to False")

        if not self.is_starter_module and not self.synthesis_domain:
            if self.type.name == 'NRPS':
                self.synthesis_domain = SynthesisDomain('C')
            elif self.type.name == 'PKS':
                self.synthesis_domain = SynthesisDomain('KS')

        if self.is_termination_module and not self.termination_domain:
            self.termination_domain = TerminationDomain("DUMMY_TE")

    def run_module(self, structure: Union[Structure, None] = None):
        raise NotImplementedError

    def get_tailoring_domain(self, domain_name: str) -> Union[TailoringDomain, None]:
        for domain in self.tailoring_domains:
            if domain.type.name == domain_name:
                return domain

        return

    def do_pks_tailoring(self, structure: Structure) -> Structure:
        kr_domain = self.get_tailoring_domain("KR")
        dh_domain = self.get_tailoring_domain("DH")
        er_domain = self.get_tailoring_domain("ER")

        if kr_domain and kr_domain.active and kr_domain.used:
            assert kr_domain.subtype is not None
            structure, kr_tailored = kr_domain.do_tailoring(structure)
            if kr_tailored:
                if not kr_domain.subtype.name == 'C1' and not kr_domain.subtype.name == 'C2':
                    if dh_domain and dh_domain.active and dh_domain.used:
                        structure, dh_tailored = dh_domain.do_tailoring(structure)
                        if dh_tailored:
                            if er_domain and er_domain.active and er_domain.used:
                                structure, er_tailored = er_domain.do_tailoring(structure)
                                if not er_tailored:
                                    er_domain.used = False

                        else:
                            dh_domain.used = False
                            if er_domain:
                                er_domain.used = False
            else:
                kr_domain.used = False
                if dh_domain:
                    dh_domain.used = False
                if er_domain:
                    er_domain.used = False
        else:
            if dh_domain:
                dh_domain.used = False
            if er_domain:
                er_domain.used = False

        return structure

    def do_nrps_tailoring(self, structure: Structure) -> Structure:
        e_domain = self.get_tailoring_domain('E')
        n_mt_domain = self.get_tailoring_domain('nMT')

        if e_domain and e_domain.active and e_domain.used:
            structure, epimerized = e_domain.do_tailoring(structure)
            if not epimerized:
                e_domain.used = False
        if n_mt_domain and n_mt_domain.active and n_mt_domain.used:
            structure, methylated = n_mt_domain.do_tailoring(structure)
            if not methylated:
                n_mt_domain.used = False

        return structure

    def release_chain(self, structure: Structure) -> Structure:
        assert structure
        assert self.termination_domain

        released_structure = self.termination_domain.release_chain(structure)

        return released_structure


class LinearPKSModule(_Module):

    def __init__(self, nr, domains, starter=False, terminator=False):
        super().__init__(nr, "PKS", domains, starter=starter, terminator=terminator, module_subtype="PKS_CIS")

    def run_module(self, structure: Union[Structure, None] = None) -> Structure:
        if structure is None:
            assert self.is_starter_module
            starter_unit = read_smiles(self.recognition_domain.substrate.smiles)
            label_pk_central_chain(starter_unit)
            structure = attach_to_domain_pk(starter_unit)
        else:
            structure = self.synthesis_domain.do_elongation(structure, self.recognition_domain.substrate)

        structure = self.do_pks_tailoring(structure)
        # RaichuDrawer(structure).draw_structure()
        return structure


class IterativePKSModule(_Module):
    def __init__(self, nr, domains, starter=False, terminator=False, iterations: int = 1) -> None:
        super().__init__(nr, "PKS", domains, starter=starter, terminator=terminator, module_subtype="PKS_ITER")
        self.iterations = iterations

    def run_module(self, structure=None):
        self.do_pks_tailoring(structure)
        raise NotImplementedError


class TransATPKSModule(_Module):
    def __init__(self, nr, domains, substrate_name="MALONYL_COA", starter=False, terminator=False) -> None:
        super().__init__(nr, "PKS", domains, starter=starter, terminator=terminator, module_subtype="PKS_TRANS")
        if not self.recognition_domain:
            self.recognition_domain = RecognitionDomain("DUMMY_AT", substrate_name)

    def do_pks_tailoring(self, structure: Structure) -> Structure:
        kr_domain = self.get_tailoring_domain("KR")
        if not kr_domain:
            kr_domain = self.get_tailoring_domain("DUMMY_KR")
        dh_domain = self.get_tailoring_domain("DH")
        if not dh_domain:
            dh_domain = self.get_tailoring_domain("DUMMY_DH")
        edh_domain = self.get_tailoring_domain("EDH")
        if not edh_domain:
            edh_domain = self.get_tailoring_domain("DUMMY_EDH")
        zdh_domain = self.get_tailoring_domain("ZDH")
        if not zdh_domain:
            zdh_domain = self.get_tailoring_domain("DUMMY_ZDH")
        er_domain = self.get_tailoring_domain("ER")
        if not er_domain:
            er_domain = self.get_tailoring_domain("DUMMY_ER")
        almt_domain = self.get_tailoring_domain("ALMT")
        if not almt_domain:
            almt_domain = self.get_tailoring_domain("DUMMY_ALMT")
        amt_domain = self.get_tailoring_domain("AMT")
        if not amt_domain:
            amt_domain = self.get_tailoring_domain("DUMMY_AMT")
        sc_domain = self.get_tailoring_domain("SC")
        if not sc_domain:
            sc_domain = self.get_tailoring_domain("DUMMY_SC")
        ah_domain = self.get_tailoring_domain("AH")
        if not ah_domain:
            ah_domain = self.get_tailoring_domain("DUMMY_AH")
        gdh_domain = self.get_tailoring_domain("GDH")
        if not gdh_domain:
            gdh_domain = self.get_tailoring_domain("DUMMY_GDH")
        zgdh_domain = self.get_tailoring_domain("ZGDH")
        if not zgdh_domain:
            zgdh_domain = self.get_tailoring_domain("DUMMY_ZGDH")
        egdh_domain = self.get_tailoring_domain("EGDH")
        if not egdh_domain:
            egdh_domain = self.get_tailoring_domain("DUMMY_EGDH")
        omt_domain = self.get_tailoring_domain("OMT")
        if not omt_domain:
            omt_domain = self.get_tailoring_domain("DUMMY_OMT")
        bmt_domain = self.get_tailoring_domain("BMT")
        if not bmt_domain:
            bmt_domain = self.get_tailoring_domain("DUMMY_BMT")

        if kr_domain and kr_domain.active and kr_domain.used:
            assert kr_domain.subtype is not None

            structure, kr_tailored = kr_domain.do_tailoring(structure)
            if not kr_tailored or kr_domain.subtype.name == 'C1' or kr_domain.subtype.name == 'C2':
                kr_domain.used = False
                if omt_domain:
                    omt_domain.used = False
                if dh_domain:
                    dh_domain.used = False
                if bmt_domain:
                    bmt_domain.used = False
                if er_domain:
                    er_domain.used = False
            if not kr_domain.subtype.name == 'C1' and not kr_domain.subtype.name == 'C2':
                if omt_domain and omt_domain.active and omt_domain.used:
                    structure, omt_tailored = omt_domain.do_tailoring(structure)
                    if not omt_tailored:
                        omt_domain.used = False
                elif dh_domain and dh_domain.active and dh_domain.used:
                    structure, dh_tailored = dh_domain.do_tailoring(structure)
                    if not dh_tailored:
                        dh_domain.used = False
                        if bmt_domain:
                            bmt_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                                if bmt_domain:
                                    bmt_domain.used = False
                            else:
                                if bmt_domain and bmt_domain.active and bmt_domain.used:
                                    structure, bmt_tailored = bmt_domain.do_tailoring(structure)
                                    if not bmt_tailored:
                                        bmt_domain.used = False
                elif edh_domain and edh_domain.active and edh_domain.used:
                    structure, edh_tailored = edh_domain.do_tailoring(structure)
                    if not edh_tailored:
                        edh_domain.used = False
                        if bmt_domain:
                            bmt_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                                if bmt_domain:
                                    bmt_domain.used = False
                            else:
                                if bmt_domain and bmt_domain.active and bmt_domain.used:
                                    structure, bmt_tailored = bmt_domain.do_tailoring(structure)
                                    if not bmt_tailored:
                                        bmt_domain.used = False
                elif zdh_domain and zdh_domain.active and zdh_domain.used:
                    structure, zdh_tailored = zdh_domain.do_tailoring(structure)
                    if not zdh_tailored:
                        zdh_domain.used = False
                        if bmt_domain:
                            bmt_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                                if bmt_domain:
                                    bmt_domain.used = False
                            else:
                                if bmt_domain and bmt_domain.active and bmt_domain.used:
                                    structure, bmt_tailored = bmt_domain.do_tailoring(structure)
                                    if not bmt_tailored:
                                        bmt_domain.used = False
                elif gdh_domain and gdh_domain.active and gdh_domain.used:
                    structure, gdh_domain_tailored = gdh_domain.do_tailoring(structure)
                    if not gdh_domain_tailored:
                        gdh_domain.used = False
                        if bmt_domain:
                            bmt_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                                if bmt_domain:
                                    bmt_domain.used = False
                            else:
                                if bmt_domain and bmt_domain.active and bmt_domain.used:
                                    structure, bmt_tailored = bmt_domain.do_tailoring(structure)
                                    if not bmt_tailored:
                                        bmt_domain.used = False
                elif egdh_domain and egdh_domain.active and egdh_domain.used:
                    structure, egdh_domain_tailored = egdh_domain.do_tailoring(structure)
                    if not egdh_domain_tailored:
                        egdh_domain.used = False
                        if bmt_domain:
                            bmt_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                                if bmt_domain:
                                    bmt_domain.used = False
                            else:
                                if bmt_domain and bmt_domain.active and bmt_domain.used:
                                    structure, bmt_tailored = bmt_domain.do_tailoring(structure)
                                    if not bmt_tailored:
                                        bmt_domain.used = False
                elif zgdh_domain and zgdh_domain.active and zgdh_domain.used:
                    structure, zgdh_domain_tailored = zgdh_domain.do_tailoring(structure)
                    if not zgdh_domain_tailored:
                        zgdh_domain.used = False
                        if bmt_domain:
                            bmt_domain.used = False
                        if er_domain:
                            er_domain.used = False
                    else:
                        if er_domain and er_domain.active and er_domain.used:
                            structure, er_tailored = er_domain.do_tailoring(structure)
                            if not er_tailored:
                                er_domain.used = False
                                if bmt_domain:
                                    bmt_domain.used = False
                            else:
                                if bmt_domain and bmt_domain.active and bmt_domain.used:
                                    structure, bmt_tailored = bmt_domain.do_tailoring(structure)
                                    if not bmt_tailored:
                                        bmt_domain.used = False
        if amt_domain and amt_domain.active and amt_domain.used:
            structure, amt_tailored = amt_domain.do_tailoring(structure)
            if not amt_tailored:
                amt_domain.used = False
        if almt_domain and almt_domain.active and almt_domain.used:
            structure, almt_tailored = almt_domain.do_tailoring(structure)
            if not almt_tailored:
                almt_domain.used = False
        if sc_domain and sc_domain.active and sc_domain.used:
            structure, sc_tailored = sc_domain.do_tailoring(structure)
            if not sc_tailored:
                sc_domain.used = False
        if ah_domain and ah_domain.active and ah_domain.used:
            structure, ah_tailored = ah_domain.do_tailoring(structure)
            if not ah_tailored:
                ah_domain.used = False
        return structure

    def run_module(self, structure: Union[Structure, None] = None) -> Structure:
        if structure is None:
            assert self.is_starter_module
            starter_unit = read_smiles(self.recognition_domain.substrate.smiles)
            label_pk_central_chain(starter_unit)
            structure = attach_to_domain_pk(starter_unit)
        else:
            structure = self.synthesis_domain.do_elongation(structure, self.recognition_domain.substrate)
        structure = self.do_pks_tailoring(structure)
        return structure


class NRPSModule(_Module):

    def __init__(self, nr, domains, starter=False, terminator=False):
        super().__init__(nr, "NRPS", domains, starter=starter, terminator=terminator)

    def run_module(self, structure=None) -> Structure:
        if structure is None:
            assert self.is_starter_module
            starter_unit = read_smiles(self.recognition_domain.substrate.smiles)
            label_nrp_central_chain(starter_unit)
            structure = attach_to_domain_nrp(starter_unit)
        else:
            structure = self.synthesis_domain.do_elongation(structure, self.recognition_domain.substrate)

        structure = self.do_nrps_tailoring(structure)
        return structure



import pygame
from interactive.style import BLACK, BUTTON_PANEL_COLOUR, FONT, WHITE


class TextBox:

    def __init__(self, position, dimensions):
        self.x, self.y = position
        self.width, self.height = dimensions
        self.rectangle = pygame.Rect(position, dimensions)
        self.text = ''

        self.font_size = self.height - 10

        self.text_x = self.x + 2
        self.text_y = self.y + int((self.height - self.font_size) / 2)

        self.font = pygame.font.SysFont(FONT, self.font_size, bold=False)

    def add_character(self, character):
        self.text += character

    def remove_character(self):
        self.text = self.text[:-1]

    def draw(self, screen):
        pygame.draw.rect(screen, WHITE, self.rectangle)
        pygame.draw.rect(screen, BLACK, self.rectangle, 1)
        rendered_text = self.font.render(self.text, True, BLACK)
        screen.blit(rendered_text, (self.text_x, self.text_y))

    def erase(self, screen):
        pygame.draw.rect(screen, BUTTON_PANEL_COLOUR, self.rectangle)
        pygame.draw.rect(screen, BUTTON_PANEL_COLOUR, self.rectangle, 1)



import os
import interactive.images.domains
from interactive.style import DOMAIN_SIZE, MODULE_SPACING, FONT, BLACK, RED

import pygame

DOMAIN_IMAGE_DIR = os.path.dirname(interactive.images.domains.__file__)


class Domain:
    def __init__(self, screen, domain_type, module, domain_id, domain_label=None):
        self.screen = screen
        self.module = module
        self.type = domain_type
        self.image = os.path.join(DOMAIN_IMAGE_DIR, f"{self.type.lower()}.png")
        self.highlight_image = os.path.join(DOMAIN_IMAGE_DIR, f"{self.type.lower()}_highlight.png")
        self.subtype = None
        self.id = domain_id
        self.rectangle = None
        self.substrate_rectangle = None
        self.substrate = None
        self.domain_label = domain_label

        self.x = 0
        self.y = 0

        self.width = DOMAIN_SIZE
        self.height = DOMAIN_SIZE

        self.font_size = self.height - 16
        self.font = pygame.font.SysFont(FONT, self.font_size, bold=False)

        self.set_rectangle()

        self.selected = False

    def set_rectangle(self):

        self.x = self.module.x + MODULE_SPACING / 2 + self.id * DOMAIN_SIZE
        self.y = self.module.y
        self.rectangle = pygame.Rect(self.x, self.y, self.width, self.height)

    def draw_domain_subtype(self):
        rendered_text = self.font.render(self.subtype, False, BLACK)
        text_rectangle = rendered_text.get_rect(center=(self.x + self.width / 2, self.y - 10))
        self.screen.blit(rendered_text, text_rectangle)

    def draw_substrate_text(self):
        colour = BLACK
        if self.module.gene.gene_number != 0 or self.module.id != 0:
            if self.substrate.starter:
                colour = RED

        rendered_text = self.font.render(self.substrate.abbr, False, colour)
        text_rectangle = rendered_text.get_rect(center=(self.x + self.width/2, self.y - 10))
        self.screen.blit(rendered_text, text_rectangle)

    def set_domain_subtype(self, subtype):
        self.subtype = subtype

    def draw(self, mouse):
        self.set_rectangle()

        if not self.selected and not self.rectangle.collidepoint(mouse):
            domain_image = pygame.image.load(self.image)
        else:
            domain_image = pygame.image.load(self.highlight_image)

        domain_image_scaled = pygame.transform.smoothscale(domain_image, (DOMAIN_SIZE, DOMAIN_SIZE))
        self.screen.blit(domain_image_scaled, self.rectangle)
        if self.substrate:
            self.draw_substrate_text()
        if self.subtype:
            self.draw_domain_subtype()






def parse_smiles(paras_file):
    name_to_smiles = {}
    with open(paras_file, 'r') as smiles_file:
        for line in smiles_file:
            line = line.strip()
            if line:
                name, smiles = line.split()
                name_to_smiles[name] = smiles

    return name_to_smiles





from pikachu.chem.structure import *
from pikachu.chem.atom import AtomAnnotations, AtomDrawProperties, Atom
from raichu.data.attributes import ATTRIBUTES


class CarrierDomain(Atom):
    def __init__(self, atom_type, atom_nr, chiral, charge, aromatic):
        super().__init__(atom_type, atom_nr, chiral, charge, aromatic)
        self.type = atom_type
        self.domain_type = None
        self.domain_nr = atom_nr
        self.annotations = AtomAnnotations()
        for annotation in ATTRIBUTES:
            self.annotations.add_annotation(annotation, default=False)

    def __repr__(self):
        return f'{self.annotations.domain_type}_{self.domain_nr}'

    def __hash__(self):
        return self.nr

    def __eq__(self, domain):
        if type(domain) == CarrierDomain:
            return self.nr == domain.nr
        else:
            return False

    def print_domaintype(self):
        print(self.domain_type)

    def copy(self):
        domain_copy = CarrierDomain(self.type, self.nr, self.chiral, self.charge, self.aromatic)
        domain_copy.hybridisation = self.hybridisation
        domain_copy.pyrrole = self.pyrrole
        domain_copy.furan = self.furan
        domain_copy.thiophene = self.thiophene
        domain_copy.draw = AtomDrawProperties()
        domain_copy.draw.colour = self.draw.colour
        domain_copy.annotations = self.annotations.copy()
        domain_copy.domain_nr = domain_copy.nr
        domain_copy.domain_type = self.domain_type
        connectivity = []

        for connection in self.connectivity:
            connectivity.append(connection)

        domain_copy.connectivity = tuple(connectivity)

        for neighbour in self.neighbours:
            domain_copy.neighbours.append(neighbour)

        # Note: when copying a structure, neighbours need to be refreshed again
        # after all atoms have been changed!

        return domain_copy

    def set_domain_type(self, domain_type):
        assert domain_type == 'ACP' or domain_type == 'AT' or\
               domain_type == 'KS' or domain_type == 'KR' or\
               domain_type == 'DH' or domain_type == 'ER' or\
               domain_type == 'TE' or domain_type == 'PCP'
        self.domain_type = domain_type
        self.annotations.set_annotation('domain_type', domain_type)


def make_domain(domain_type, next_atom_nr):
    """
    Returns a Domain of the input type containing the appropriate nr in the
    Structure object

    domain_type: Str, domain type
    next_atom_nr: Int, 'atom nr' of the domain in the structure
    """
    domain = CarrierDomain('I', next_atom_nr, None, 0, False)
    domain.set_domain_type(domain_type)
    return domain


def make_scaffold_domain(domain_type):
    """
    Returns a Domain of the input type containing the appropriate nr in the
    Structure object

    domain_type: Str, domain type
    next_atom_nr: Int, 'atom nr' of the domain in the structure
    """
    structure = Structure()

    domain = CarrierDomain('I', 0, None, 0, False)
    domain.set_domain_type(domain_type)

    structure.add_disconnected_atom(domain)

    sulphur = Atom('S', 1, None, 0, False)
    structure.add_bond(domain, sulphur, 'single', 0)
    structure.refine_structure()

    return structure







from typing import List

from pikachu.drawing.drawing import Drawer

from raichu.reactions.chain_release import cyclic_release
from raichu.drawing.drawer import RaichuDrawer
from raichu.module import _Module

from raichu.drawing.bubbles import draw_bubbles


from raichu.substrate import PKSSubstrate
from raichu.data.trans_at import TRANSATOR_CLADE_TO_TAILORING_REACTIONS, TRANSATOR_CLADE_TO_STARTER_SUBSTRATE, \
    TRANSATOR_CLADE_TO_ELONGATING
from raichu.domain.domain import TailoringDomain


class Cluster:
    def __init__(self, modules: List[_Module]) -> None:
        self.modules = modules
        self.chain_intermediate = None

        self.structure_intermediates = []
        self.linear_product = None
        self.cyclised_products = []
        self.module_mechanisms = []
        self.handle_transat()

    def handle_transat(self):
        for i, module in enumerate(self.modules):
            if module.type.name == "PKS" and module.subtype.name == "PKS_TRANS":
                substrate = PKSSubstrate("MALONYL_COA")
                if i < len(self.modules) - 1:
                    next_module = self.modules[i + 1]
                    if next_module.type.name == "PKS":
                        if module.is_starter_module and next_module.subtype.name == "PKS_TRANS":
                            substrate_name = TRANSATOR_CLADE_TO_STARTER_SUBSTRATE.get(next_module.synthesis_domain.subtype.name)
                            if substrate_name is not None:
                                substrate = PKSSubstrate(substrate_name)
                            else:
                                substrate = PKSSubstrate("ACETYL_COA")
                                # TODO: Transfer names of substrates to a dictionary in raichu.substrate
                        elif module.is_starter_module:
                            substrate = PKSSubstrate("ACETYL_COA")
                        if not module.is_termination_module and next_module.subtype.name == "PKS_TRANS":
                            # Ignore tailoring domains in the module itself
                            # Why do we do this if we delete them afterwards anyways?
                            for domain in module.tailoring_domains:
                                domain.used = False
                            module.tailoring_domains = []
                            for dummy_domain_type, dummy_domain_subtype in TRANSATOR_CLADE_TO_TAILORING_REACTIONS[next_module.synthesis_domain.subtype.name]:
                                self.modules[i].tailoring_domains.append(TailoringDomain(dummy_domain_type,
                                                                                         dummy_domain_subtype))
                if module.synthesis_domain:
                    if not TRANSATOR_CLADE_TO_ELONGATING[module.synthesis_domain.subtype.name]:
                        self.modules[i].synthesis_domain.is_elongating = False
                self.modules[i].recognition_domain.substrate = substrate

    def compute_structures(self, compute_cyclic_products=True):
        for module in self.modules:
            structure = module.run_module(self.chain_intermediate)
            self.structure_intermediates.append(structure.deepcopy())
            self.chain_intermediate = structure
            if module.is_termination_module:
                self.linear_product = module.release_chain(structure)
                break
        else:
            raise ValueError("Cluster must contain at least one termination module.")

        if compute_cyclic_products:
            self.cyclise_all()

    def cyclise(self, atom):
        cyclic_release(self.linear_product, atom)

    def cyclise_all(self):
        pass

    def draw_spaghettis(self):
        spaghetti_svgs = []
        for structure in self.structure_intermediates:
            drawing = RaichuDrawer(structure, dont_show=True)
            drawing.draw_structure()
            svg_string = drawing.save_svg_string()
            spaghetti_svgs.append(svg_string)

        linear_drawing = Drawer(self.linear_product)
        linear_svg = linear_drawing.save_svg_string()

        return spaghetti_svgs + [linear_svg]

    def get_drawings(self, whitespace=30):

        drawings = []
        widths = []

        for i, structure in enumerate(self.structure_intermediates):

            drawing = RaichuDrawer(structure, dont_show=True)
            drawing.flip_y_axis()
            drawing.move_to_positive_coords()
            drawing.convert_to_int()

            carrier_domain_pos = None

            for atom in drawing.structure.graph:
                if atom.annotations.domain_type:
                    carrier_domain_pos = atom.draw.position
                    atom.draw.positioned = False

            assert carrier_domain_pos

            min_x = 100000000
            max_x = -100000000

            for atom in drawing.structure.graph:
                if atom.draw.positioned:
                    if atom.draw.position.x < min_x:
                        min_x = atom.draw.position.x
                    if atom.draw.position.x > max_x:
                        max_x = atom.draw.position.x

            width = (carrier_domain_pos.x - min_x + 0.5 * whitespace,
                     max_x - carrier_domain_pos.x + 0.5 * whitespace)
            widths.append(width)
            drawings.append(drawing)

        return drawings, widths

    def draw_cluster(self):
        drawings, widths = self.get_drawings()
        bubble_svg, bubble_positions, last_domain_coord = draw_bubbles(self, widths)
        min_x = 100000000
        max_x = -100000000
        min_y = 100000000
        max_y = -100000000

        svg_strings = []
        squiggly_svgs = []
        padding = None
        svg_style = None

        for i, drawing in enumerate(drawings):
            drawing.set_structure_id(f"s{i}")
            svg_style = drawing.svg_style

            padding = drawing.options.padding

            carrier_domain_pos = None

            for atom in drawing.structure.graph:
                if atom.annotations.domain_type:
                    carrier_domain_pos = atom.draw.position
                    atom.draw.positioned = False

            assert carrier_domain_pos
            bubble_x, bubble_y = bubble_positions[i]
            x_translation = bubble_x - carrier_domain_pos.x
            y_translation = bubble_y - carrier_domain_pos.y

            drawing.move_structure(x_translation, y_translation + 15)
            svg = drawing.draw_svg()
            svg_strings.append(svg)

            sulphur_pos = None
            carrier_pos = None

            for atom in drawing.structure.graph:
                if atom.draw.positioned:
                    if atom.draw.position.x < min_x:
                        min_x = atom.draw.position.x
                    if atom.draw.position.y < min_y:
                        min_y = atom.draw.position.y
                    if atom.draw.position.x > max_x:
                        max_x = atom.draw.position.x
                    if atom.draw.position.y > max_y:
                        max_y = atom.draw.position.y
                if atom.annotations.domain_type:
                    sulphur_pos = atom.get_neighbour('S').draw.position
                    carrier_pos = atom.draw.position

            squiggly_svg = f'<path d="M {sulphur_pos.x} {sulphur_pos.y - 5} Q {sulphur_pos.x - 5} {sulphur_pos.y - (sulphur_pos.y - 5 - carrier_pos.y)/2}, {carrier_pos.x} {sulphur_pos.y - 5 - (sulphur_pos.y - 5 - carrier_pos.y)/2} T {carrier_pos.x} {carrier_pos.y}" stroke="grey" fill="white"/>'
            squiggly_svgs.append(squiggly_svg)
        assert padding is not None

        x1 = 0
        x2 = max([max_x + padding, last_domain_coord + padding])
        y1 = 0
        y2 = max_y + padding

        width = x2
        height = y2

        svg_string = f"""<svg width="{width}" height="{height}" viewBox="{x1} {y1} {x2} {y2}" xmlns="http://www.w3.org/2000/svg">\n"""
        if svg_style:
            svg_string += f"{svg_style}\n"
        svg_string += bubble_svg
        for string in svg_strings:
            svg_string += string
        for squiggly_svg in squiggly_svgs:
            svg_string += squiggly_svg
        svg_string += "</svg>"

        with open('testy.svg', 'w') as svg_out:
            svg_out.write(svg_string)


class Mechanism:
    def __init__(self):
        self.structures = []

    def add_structure(self, structure):
        self.structures.append(structure.deepcopy())

    def draw_mechanism(self):
        pass



import os

from pikachu.general import read_smiles, draw_structure
from pikachu.reactions.functional_groups import BondDefiner, GroupDefiner
from raichu.reactions.general import initialise_atom_attributes

import raichu.data

FLATFILES = os.path.dirname(raichu.data.__file__)
AA_SMILES = os.path.join(FLATFILES, "PARAS_smiles.txt")


class PksElongationUnit:
    def __init__(self, name, smiles, c_idx, carbonyl_idx):
        """
        name: str, elongation unit name
        smiles: str, elongation unit SMILES
        c_idx: int, index of carbon that will be attached to the thioester carbon of the chain intermediate
        carbonyl_idx: int, index of the carbon of the carbonyl group
        """
        self.name = name
        self.smiles = smiles
        self.structure = read_smiles(self.smiles)
        initialise_atom_attributes(self.structure)
        self.structure.refresh_structure()

        self.c_to_pk_intermediate = None
        self.c_to_s = None

        for atom in self.structure.graph:
            if atom.nr == c_idx:
                self.c_to_pk_intermediate = atom
            elif atom.nr == carbonyl_idx:
                self.c_to_s = atom

        assert self.c_to_pk_intermediate is not None and self.c_to_s is not None

        self.c_to_pk_intermediate.annotations.in_central_chain = True
        self.c_to_s.annotations.in_central_chain = True

        if self.smiles == 'O=CCC':
            # TODO: CHECK CHIRALITY ACTUALLY WORKS!!!!
            self.c_to_pk_intermediate.chiral = 'clockwise'


def parse_smiles():
    """
    Returns a dict as {name_aa : SMILES_aa} from the amino acids in the
    PARAS_smiles.txt file
    """
    # Parse list SMILES amino acids attached to PCP to dict name -> Structure
    name_to_smiles = {}
    with open(AA_SMILES, 'r', encoding='utf8') as paras_smiles:
        for line in paras_smiles:
            line = line.strip()
            print(line)
            name, smiles = line.split('\t')
            name = name.upper()
            name_to_smiles[name] = smiles

    return name_to_smiles


def make_elongation_monomer(name):
    if name == 'MALONYL_COA':
        monomer = PksElongationUnit('Malonyl CoA', 'CC=O', 0, 1)
    elif name == 'METHYLMALONYL_COA':
        monomer = PksElongationUnit('Methylmalonyl CoA', 'O=CCC', 2, 1)
    elif name == 'METHOXYMALONYL_ACP':
        monomer = PksElongationUnit("Methoxymalonyl ACP", 'O=CCOC', 2, 1)
    elif name == 'ETHYLMALONYL_COA':
        monomer = PksElongationUnit("Ethylmalonyl CoA", 'O=CCCC', 2, 1)
    elif name == 'WILDCARD':
        monomer = PksElongationUnit("Wildcard", 'O=CC*', 2, 1)
    else:
        raise ValueError(f"{name} not recognised by RAIChU as PKS elongation unit.")

    return monomer


THIOESTERBOND = BondDefiner('thioester_bond', 'SC(C)=O', 0, 1)
THIOESTER_CARBON = GroupDefiner('thioester carbon', 'SC(C)=O', 1)
LEAVING_OH_BOND = BondDefiner('Leaving -OH group bond', 'C(=O)(O)CN', 0, 2)
N_AMINO_ACID = GroupDefiner('Nitrogen atom amino acid', 'NCC(=O)O', 0)
C1_AMINO_ACID = GroupDefiner('C1 atom amino acid', 'NCC(=O)O', 1)
C2_AMINO_ACID = GroupDefiner('C2 atom amino acid', 'NCC(=O)O', 2)
AMINO_ACID_BACKBONE = read_smiles('NCC(=O)O')
BETA_AMINO_ACID_BACKBONE = read_smiles('NCCC(=O)O')
B_N_AMINO_ACID = GroupDefiner('Nitrogen atom beta amino acid', 'NCCC(=O)O', 0)
B_C1_AMINO_ACID = GroupDefiner('C1 atom beta amino acid', 'NCCC(=O)O', 1)
B_C2_AMINO_ACID = GroupDefiner('C2 atom beta amino acid', 'NCCC(=O)O', 2)
B_C3_AMINO_ACID = GroupDefiner('C3 atom beta amino acid', 'NCCC(=O)O', 3)
B_LEAVING_BOND = BondDefiner('beta leaving bond', 'NCCC(=O)O', 3, 5)
AMINO_FATTY_ACID = read_smiles('CCCCCCCC(=O)O')
ACID_C1 = GroupDefiner('C1 atom (fatty) acid', 'CC(=O)O', 0)
ACID_C2 = GroupDefiner('C2 atom (fatty) acid', 'CC(=O)O', 1)

COABOND = BondDefiner('CoA_bond', 'CC(NCCC(NCCSC)=O)=O', 8, 9)

SC_BOND = BondDefiner('recent_elongation', 'SC(C)=O', 0, 1)
CO_BOND = BondDefiner('recent_elongation', 'CO', 0, 1)
N_AMINO = GroupDefiner('N_amino', 'CN', 1)
O_OH = GroupDefiner('O_oh', 'CO', 1)

O_BETAPROPRIOLACTONE = GroupDefiner('o_betapropriolactone', 'SC(CCO)=O', 4)
O_BETAPROPRIOLACTONE_O = GroupDefiner('o_betapropriolactone', 'OC(CCO)=O', 4)
O_BETAPROPRIOLACTONE_TERMINAL_O = GroupDefiner('o_betapropriolactone', 'OC(CCO)=O', 0)



from raichu.class_domain import *
from pikachu.reactions.functional_groups import find_atoms, GroupDefiner, combine_structures
from pikachu.reactions.basic_reactions import condensation
from raichu.reactions.general import initialise_atom_attributes
from pikachu.general import read_smiles

POLYKETIDE_S = GroupDefiner('Sulphur atom polyketide', 'SC(C)=O', 0)
NRP_C = GroupDefiner('C atom to attach to PCP domain', 'NCC(O)=O', 2)
AMINO_ACID_BACKBONE = read_smiles('NCC(=O)O')
B_AMINO_ACID_BACKBONE = read_smiles('NCCC(=O)O')
ASP_ACID_C = GroupDefiner('Asp', 'NC(C=O)CC(O)=O', 5)
B_NRP_C = GroupDefiner('Beta C atom to attach to PCP domain', 'NCCC(O)=O', 3)
MAL_AMINO = GroupDefiner('malonyl_starter_amino', 'NC(=O)CC(O)=O', 4)


def attach_to_domain_pk(polyketide):
    """
    Attaches the sulphur atom in the input polyketide to a PKS domain and
    returns the attached structure as a PIKAChU Structure object

    polyketide: PIKAChU Structure object, to-be attached structure
    """
    # Add attributes to input molecule
    initialise_atom_attributes(polyketide)

    # Create domain
    domain = make_scaffold_domain('ACP')
    sh_bond = domain.bond_lookup[domain.atoms[1]][domain.atoms[2]]
    hydrogen = domain.atoms[2]
    sulphur_1 = domain.atoms[1]

    locations_sulphur = find_atoms(POLYKETIDE_S, polyketide)
    assert len(locations_sulphur) == 1
    sulphur_2 = locations_sulphur[0]
    carbon = sulphur_2.get_neighbour('C')
    sc_bond = sulphur_2.get_bond(carbon)

    structure = combine_structures([domain, polyketide])
    structure.break_bond(sh_bond)
    structure.break_bond(sc_bond)

    structure.make_bond(hydrogen, sulphur_2, structure.find_next_bond_nr())
    structure.make_bond(carbon, sulphur_1, structure.find_next_bond_nr())

    split = structure.split_disconnected_structures()

    tethered_polyketide = None

    for structure in split:
        if carbon in structure.graph:
            tethered_polyketide = structure
            break

    assert tethered_polyketide
    initialise_atom_attributes(tethered_polyketide)

    return tethered_polyketide


def attach_to_domain_nrp(nrp):
    """
    Attaches the input NRP to a PCP domain and returns the product as a
    PIKAChU Structure object

    nrp: PIKAChU Structure object, to-be attached NRP
    """

    initialise_atom_attributes(nrp)

    # Create domain
    domain = make_scaffold_domain('PCP')

    # Remove OH group from carboxylic acid in NRP, to allow attachment
    # to domain

    if nrp.find_substructures(AMINO_ACID_BACKBONE):
        locations_c_to_domain = find_atoms(NRP_C, nrp)
        assert len(locations_c_to_domain) == 1
        c_atom_to_domain = locations_c_to_domain[0]
    elif nrp.find_substructures(B_AMINO_ACID_BACKBONE):
        locations_c_to_domain = find_atoms(B_NRP_C, nrp)
        asp_acid_cs = find_atoms(ASP_ACID_C, nrp)
        mal_amino_cs = find_atoms(MAL_AMINO, nrp)

        for asp_acid_c in asp_acid_cs:
            if asp_acid_c in locations_c_to_domain:
                locations_c_to_domain.remove(asp_acid_c)
        for mal_amino_c in mal_amino_cs:
            if mal_amino_c in locations_c_to_domain:
                locations_c_to_domain.remove(mal_amino_c)
        assert len(locations_c_to_domain) == 1
        c_atom_to_domain = locations_c_to_domain[0]
    else:
        c_atom_to_domain = None
        for atom in nrp.graph:
            if atom.annotations.c2_acid:
                c_atom_to_domain = atom

    assert c_atom_to_domain

    oxygens = c_atom_to_domain.get_neighbours('O')

    hydroxyl_oxygen = None
    hydroxyl_bond = None

    hydrogen_bond = domain.bond_lookup[domain.atoms[1]][domain.atoms[2]]

    assert hydrogen_bond.has_neighbour('S')
    assert hydrogen_bond.has_neighbour('H')

    for oxygen in oxygens:
        bond = c_atom_to_domain.get_bond(oxygen)
        if bond.type == 'single' and oxygen.has_neighbour('H'):
            hydroxyl_oxygen = oxygen
            hydroxyl_bond = bond
            break

    assert hydroxyl_oxygen and hydroxyl_bond

    structure = condensation(nrp, domain, hydroxyl_bond, hydrogen_bond)[0]

    initialise_atom_attributes(structure)

    return structure



TRANSATOR_CLADE_TO_TAILORING_REACTIONS = {'NON_ELONGATING_DB': [('DUMMY_KR', None), ('DUMMY_DH', None)],
                                          'BETA_OH_KETO': [],
                                          'NON_ELONGATING_BETA_OH': [('DUMMY_KR', None)],
                                          'ALPHA_OH': [('DUMMY_AH', None)],
                                          'EDB': [('DUMMY_KR', None), ('DUMMY_EDH', None)],
                                          'ARST': [],
                                          'PYR': [('DUMMY_SC', None)],
                                          'ALPHAME_EDB': [('DUMMY_AMT', None), ('DUMMY_KR', None),
                                                                       ('DUMMY_EDH', None)],
                                          'OXI': [], 'AA': [],
                                          'ALPHAME_BETA_L_OH': [('DUMMY_AMT', None),
                                                                             ('DUMMY_KR', 'B1')],
                                          'NON_ELONGATING_BETA_L_OH': [('DUMMY_KR', 'B1')],
                                          'RED_SHDB': [('DUMMY_KR', None), ('DUMMY_GDH', None)],
                                          'DB': [('DUMMY_KR', None), ('DUMMY_DH', None)],
                                          'KETO': [],
                                          'NON_ELONGATING': [],
                                          'BETA_D_OH': [('DUMMY_KR', None)],
                                          'BETA_L_OH': [('DUMMY_KR', 'B1')], 'BR': [],
                                          'ALPHABETA_OH': [('DUMMY_AH', None), ('DUMMY_KR', None)],
                                          'UNST': [],
                                          'ST': [],
                                          'MEOST': [],
                                          'ACST': [],
                                          'ALPHAME': [('DUMMY_KR', None), ('DUMMY_DH', None),
                                                                   ('DUMMY_ER', None), ('DUMMY_AMT', None)],
                                          'BETA_D_OME': [('DUMMY_KR', None), ('DUMMY_OMT', None)],
                                          'NON_ELONGATING_PYR': [('DUMMY_SC', None)],
                                          'BETA_ME': [('DUMMY_KR', None), ('DUMMY_DH', None),
                                                                   ('DUMMY_ER', None), ('DUMMY_BMT', None)],
                                          'BETA_OH_EDB': [('DUMMY_KR', None)],
                                          'NON_ELONGATING_OXA': [],
                                          'LACST': [],
                                          'SHDB': [('DUMMY_KR', None), ('DUMMY_GDH', None)],
                                          'OUT': [],
                                          'BETA_OH': [('DUMMY_KR', None)],
                                          'ZDB': [('DUMMY_KR', None), ('DUMMY_ZDH', None)],
                                          'BETA_MEDB': [('DUMMY_KR', None), ('DUMMY_EDH', None),
                                                                     ('DUMMY_BMT', None)],
                                          'TRANS_AT_PKS': [], 'OXA': [],
                                          'ALPHAME_BETA_D_OH': [('DUMMY_ALMT', None), ('DUMMY_KR', 'A1')],
                                          'ALPHAME_BETAOH': [('DUMMY_AMT', None), ('DUMMY_KR', None)],
                                          'NON_ELONGATING_ALPHAME_EDB': [('DUMMY_AMT', None),
                                                                                      ('DUMMY_KR', None),
                                                                                      ('DUMMY_EDH', None)],
                                          'RED': [('DUMMY_KR', None), ('DUMMY_DH', None),
                                                                   ('DUMMY_ER', None)]}

# TODO: Map all these SMILES to names, make Enums for them
TRANSATOR_CLADE_TO_STARTER_SUBSTRATE = {'NON_ELONGATING_DB': 'ACETYL_COA',
                                        # TODO: choose SMILES
                                        'BETA_OH_KETO': 'ACETYL_COA',
                                        'NON_ELONGATING_BETA_OH': 'ACETYL_COA',
                                        'ALPHA_OH': 'ACETYL_COA',
                                        'EDB': "ACETYL_COA",
                                        'ARST': 'PHENYLACETYLCOA', 'PYR': 'ACETYL_COA',
                                        'ALPHAME_EDB': "ACETYL_COA",
                                        'OXI': 'ACETYL_COA', 'AA': 'GLYCINE',
                                        'ALPHAME_BETA_L_OH': "ACETYL_COA",
                                        'NON_ELONGATING_BETA_L_OH': "ACETYL_COA",
                                        'RED_SHDB': 'ACETYL_COA',
                                        'DB': "ACETYL_COA",
                                        'KETO': 'ACETYL_COA',
                                        'NON_ELONGATING': 'ACETYL_COA',
                                        'BETA_D_OH': "ACETYL_COA",
                                        'BETA_L_OH': "ACETYL_COA",
                                        'BR': 'ACETYL_COA',
                                        'ALPHABETA_OH': 'ACETYL_COA',
                                        'UNST': 'PROPIONYL_COA', 'ST': 'ACETYL_COA',
                                        'MEOST': 'METHOXYFORMYL_COA',
                                        'ACST': 'ACETYL_COA',
                                        'ALPHAME': 'ACETYL_COA',
                                        'BETA_D_OME': "MACETYL_COA",
                                        'NON_ELONGATING_PYR': 'ACETYL_COA',
                                        'BETA_ME': 'ACETYL_COA',
                                        'BETA_OH_EDB': 'ACETYL_COA',
                                        'NON_ELONGATING_OXA': 'ACETYL_COA',
                                        'LACST': 'LACTYL_COA',
                                        'SHDB': 'ACETYL_COA',
                                        'OUT': 'ACETYL_COA',
                                        'BETA_OH': "ACETYL_COA",
                                        'ZDB': 'ACETYL_COA',
                                        'BETA_MEDB': "ACETYL_COA",
                                        'TRANS_AT_PKS': 'ACETYL_COA', 'OXA': 'ACETYL_COA',
                                        'ALPHAME_BETA_D_OH': 'ACETYL_COA',
                                        'ALPHAME_BETAOH': 'ACETYL_COA',
                                        'NON_ELONGATING_ALPHAME_EDB': "ACETYL_COA",
                                        'RED': 'ACETYL_COA'}

TRANSATOR_CLADE_TO_ELONGATING = {'NON_ELONGATING_DB': False,
                                 'BETA_OH_KETO': True,
                                 'NON_ELONGATING_BETA_OH': False,
                                 'ALPHA_OH': True, 'EDB': True,
                                 'ARST': True, 'PYR': True,
                                 'ALPHAME_EDB': True, 'OXI': True,
                                 'AA': True, 'ALPHAME_BETA_L_OH': True,
                                 'NON_ELONGATING_BETA_L_OH': False,
                                 'RED_SHDB': True, 'DB': True,
                                 'KETO': True, 'NON_ELONGATING': False,
                                 'BETA_D_OH': True, 'BETA_L_OH': True,
                                 'BR': True, 'ALPHABETA_OH': True,
                                 'UNST': True, 'ST': True,
                                 'MEOST': True, 'ACST': True,
                                 'ALPHAME': True, 'BETA_D_OME': True,
                                 'NON_ELONGATING_PYR': False,
                                 'BETA_ME': True, 'BETA_OH_EDB': True,
                                 'NON_ELONGATING_OXA': False,
                                 'LACST': True,
                                 'SHDB': True, 'OUT': True,
                                 'BETA_OH': True, 'ZDB': True,
                                 'BETA_MEDB': True, 'TRANS_AT_PKS': True,
                                 'OXA': True, 'ALPHAME_BETA_D_OH': True,
                                 'ALPHAME_BETAOH': True,
                                 'NON_ELONGATING_ALPHAME_EDB': False,
                                 'RED': True}



from setuptools import setup, find_packages

VERSION = '0.0.1'
DESCRIPTION = 'RAIChU'
LONG_DESCRIPTION = 'An interactive PKS/NRPS visualizer.'

print(find_packages())

setup(
    name="raichu",
    version=VERSION,
    author="Barbara Terlouw",
    author_email="barbara.r.terlouw@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(exclude="build"),
    package_data={"": ["*.png", "*.svg", "*.txt"]},
    install_requires=['matplotlib',
                      'pikachu-chem',
                      'paras'],
    scripts=["bin/raichu"],)



import os

import pygame

from interactive.buttons import make_buttons, get_mouse_button, Button, DomainButton, hide_button, \
    AddGeneButton, CreateGeneButton, show_buttons, ADD_MODULE_BUTTON, show_button, \
    AddModuleButton, NRPSModuleButton, NRPS_MODULE_BUTTON, PKSModuleButton, PKS_MODULE_BUTTON, \
    show_domain_buttons, reset_buttons, AddDomainButton, ADD_DOMAIN_BUTTON, \
    REMOVE_MODULE_BUTTON, REMOVE_DOMAIN_BUTTON, RemoveDomainButton, RemoveModuleButton, RemoveGeneButton, \
    REMOVE_GENE_BUTTON, SELECT_SUBSTRATE_BUTTON, SelectSubstrateButton, SELECT_DOMAIN_TYPE_BUTTON, \
    SelectDomainTypeButton, PKS_SUBSTRATE_BUTTONS, SubstrateSupergroupButton,\
    PROTEINOGENIC_BUTTONS, NON_PROTEINOGENIC_BUTTONS, FATTY_ACID_BUTTONS, NON_AMINO_ACID_BUTTONS, \
    ProteinogenicButton, NonProteinogenicButton, FattyAcidButton, NonAminoAcidButton, \
    PROTEINOGENIC_GROUP_TO_BUTTONS, NONPROTEINOGENIC_GROUP_TO_BUTTONS, FATTY_ACID_GROUP_TO_BUTTONS, \
    NON_AMINOACID_GROUP_TO_BUTTONS, SubstrateGroupButton, SubstrateButton, RenderClusterButton, \
    KR_BUTTONS, KRSubtypeButton, SaveClusterButton, SaveProductsButton, RenderProductsButton, \
    NRPSWildcardButton, SaveToPngButton, SaveToFolderButton, FATTY_ACID_SUPER_OPTIONS_BUTTONS, \
    hide_buttons, ALL_BUTTONS, SAVE_PRODUCTS_BUTTON, SAVE_CLUSTER_BUTTON, YES_BUTTON, NO_BUTTON, YesButton, NoButton, \
    FattyAcidOptionButton, FattyAcidSuperOptionButton, ISO_BUTTON, ANTEISO_BUTTON, show_carbon_nr_buttons, \
    CIS_BUTTON, TRANS_BUTTON, UNDEFINED_BUTTON, CButton, SET_ISOFORM_BUTTON, StarterButton, ElongationButton, \
    PKS_STARTER_SUBSTRATE_BUTTONS, PROTEINOGENIC_BUTTON, FATTY_ACID_BUTTON, NON_PROTEINOGENIC_BUTTON, \
    NON_AMINO_ACID_BUTTON, STARTER_BUTTON, ELONGATION_BUTTON, ExportTabularButton, SaveToTxtButton, NRPS_WILDCARD_BUTTON
from interactive.domain import Domain
from interactive.module import Module
from interactive.gene import Gene
from interactive.substrate import FattyAcid
from interactive.insertion_point import InsertionPoint
from interactive.render_cluster import render_cluster, render_products, export_tabular
from interactive.style import RENDER_WINDOW_SIZE, WHITE, REPLACE_TEXT_1, REPLACE_TEXT_2, HEIGHT, WIDTH, \
    SIZE, FATTY_ACID_IMAGE_SIZE, FAILED_TEXT_1, FAILED_TEXT_3, FAILED_TEXT_2


class RaichuManager:
    states = {'add gene',
              'remove_gene',
              'add module',
              'remove module',
              'add domain',
              'remove domain',
              'set domain subtype',

              }

    def __init__(self, screen):
        self.genes = []
        self.state = None
        self.active_buttons = make_buttons(screen)
        self.screen = screen

        self.selected_module = None
        self.selected_gene = None
        self.selected_domain = None
        self.selected_substrate_group = None

        self.text_box = None
        self.insertion_point = None
        self.cluster_image = None
        self.product_images = []
        self.replace_text = [REPLACE_TEXT_1, REPLACE_TEXT_2]
        self.fatty_acid_mode = 'size'
        self.fatty_acid = None
        self.fatty_acid_bond_type = None
        self.fatty_acid_modified = False

    def get_mouse_domain(self, mouse):
        for gene in self.genes:
            for module in gene.modules:
                for domain in module.domains:
                    if domain.rectangle.collidepoint(mouse):
                        return domain

        return None

    def get_mouse_module(self, mouse):
        for gene in self.genes:
            for module in gene.modules:
                if module.rectangle.collidepoint(mouse):
                    return module

        return None

    def get_mouse_gene(self, mouse):
        for gene in self.genes:
            if gene.rectangle.collidepoint(mouse):
                return gene

        return None

    def detect_click_location(self, mouse):

        button = get_mouse_button(self.active_buttons, mouse)
        domain = self.get_mouse_domain(mouse)
        module = self.get_mouse_module(mouse)
        gene = self.get_mouse_gene(mouse)

        if button:
            return button
        elif domain:
            self.reset_selections()
            domain.selected = True
            self.selected_domain = domain
            return domain
        elif module:
            self.reset_selections()
            module.selected = True
            self.selected_module = module
            return module
        elif gene:
            self.reset_selections()
            gene.selected = True
            self.selected_gene = gene
            return gene
        else:
            return None

    def reset_selections(self):
        if self.selected_gene:
            self.selected_gene.insertion_point = None
            self.selected_gene.selected = False
            self.selected_gene = None
        if self.selected_module:
            self.selected_module.selected = False
            self.selected_module = None
        if self.selected_domain:
            self.selected_domain.selected = False
            self.selected_domain = None
        self.selected_substrate_group = None

    def do_button_action(self, button, mouse):
        if type(button) == DomainButton or issubclass(type(button), DomainButton):
            button.do_action(self.selected_module, mouse, self.screen, self.active_buttons)
            self.selected_module.selected = False
            self.selected_module = None
            hide_button(button, self.screen, self.active_buttons)
        elif issubclass(type(button), SubstrateSupergroupButton):
            reset_buttons(self.screen, self.active_buttons)
            if type(button) == ProteinogenicButton:
                show_buttons(PROTEINOGENIC_BUTTONS, self.screen, self.active_buttons)
                self.selected_substrate_group = PROTEINOGENIC_GROUP_TO_BUTTONS
            elif type(button) == NonProteinogenicButton:
                show_buttons(NON_PROTEINOGENIC_BUTTONS, self.screen, self.active_buttons)
                self.selected_substrate_group = NONPROTEINOGENIC_GROUP_TO_BUTTONS
            elif type(button) == FattyAcidButton:
                show_buttons(FATTY_ACID_BUTTONS, self.screen, self.active_buttons)
                self.selected_substrate_group = FATTY_ACID_GROUP_TO_BUTTONS
            elif type(button) == NonAminoAcidButton:
                show_buttons(NON_AMINO_ACID_BUTTONS, self.screen, self.active_buttons)
                self.selected_substrate_group = NON_AMINOACID_GROUP_TO_BUTTONS
        elif type(button) == SubstrateGroupButton:
            reset_buttons(self.screen, self.active_buttons)
            if button.text != 'Custom':
                show_buttons(self.selected_substrate_group[button.text], self.screen, self.active_buttons)
                self.selected_substrate_group = None
            else:
                show_carbon_nr_buttons(self.screen, self.active_buttons)

        elif type(button) == StarterButton:
            reset_buttons(self.screen, self.active_buttons)
            if self.selected_domain.module.type == 'NRPS':
                show_buttons([FATTY_ACID_BUTTON, NON_AMINO_ACID_BUTTON, NRPS_WILDCARD_BUTTON], self.screen, self.active_buttons)
            elif self.selected_domain.module.type == 'PKS':
                show_buttons(PKS_STARTER_SUBSTRATE_BUTTONS, self.screen, self.active_buttons)

        elif type(button) == ElongationButton:
            reset_buttons(self.screen, self.active_buttons)
            if self.selected_domain.module.type == 'NRPS':
                show_buttons([PROTEINOGENIC_BUTTON, NON_PROTEINOGENIC_BUTTON, NRPS_WILDCARD_BUTTON], self.screen, self.active_buttons)
            elif self.selected_domain.module.type == 'PKS':
                show_buttons(PKS_SUBSTRATE_BUTTONS, self.screen, self.active_buttons)

        elif type(button) == FattyAcidSuperOptionButton:

            reset_buttons(self.screen, self.active_buttons)

            if button.text == 'Set substrate':
                self.fatty_acid.set_substrate(self.selected_domain)
                self.selected_domain.module.gene.erase()
                self.selected_domain.module.gene.draw(mouse)
                reset_buttons(self.screen, self.active_buttons)
                self.reset_selections()
                self.fatty_acid = None
                self.fatty_acid_mode = 'size'
                self.fatty_acid_bond_type = None
                self.fatty_acid_modified = False
                pygame.draw.rect(self.screen, WHITE,
                                 pygame.Rect(SIZE[0] - (FATTY_ACID_IMAGE_SIZE[0] + 20), SIZE[1] / 2 + 20,
                                             FATTY_ACID_IMAGE_SIZE[0], FATTY_ACID_IMAGE_SIZE[1]))
                pygame.draw.rect(self.screen, WHITE,
                                 pygame.Rect(SIZE[0] - (FATTY_ACID_IMAGE_SIZE[0] + 20), SIZE[1] / 2 + 20,
                                             FATTY_ACID_IMAGE_SIZE[0], FATTY_ACID_IMAGE_SIZE[1]), 2)

            else:

                if button.text == 'Add double bond':
                    self.fatty_acid_mode = 'double bond'
                    show_buttons([CIS_BUTTON, TRANS_BUTTON, UNDEFINED_BUTTON], self.screen, self.active_buttons)

                elif button.text == 'Set isoform':
                    if self.fatty_acid.c_nr > 4:
                        show_buttons([ISO_BUTTON, ANTEISO_BUTTON], self.screen, self.active_buttons)
                    elif self.fatty_acid.c_nr == 4:
                        show_button(ISO_BUTTON, self.screen, self.active_buttons)

                else:
                    if button.text == 'Add amino group':
                        self.fatty_acid_mode = 'amino'
                    elif button.text == 'Add methyl group':
                        self.fatty_acid_mode = 'methyl'
                    elif button.text == 'Add -OH group':
                        self.fatty_acid_mode = 'hydroxyl'

                    show_carbon_nr_buttons(self.screen, self.active_buttons, mode=self.fatty_acid_mode, fatty_acid=self.fatty_acid)
                show_buttons(FATTY_ACID_SUPER_OPTIONS_BUTTONS, self.screen, self.active_buttons)
                if self.fatty_acid_modified:
                    hide_button(SET_ISOFORM_BUTTON, self.screen, self.active_buttons)

        elif type(button) == FattyAcidOptionButton:
            reset_buttons(self.screen, self.active_buttons)
            if button.text == 'cis':
                self.fatty_acid_bond_type = 'cis'
            elif button.text == 'trans':
                self.fatty_acid_bond_type = 'trans'
            elif button.text == 'iso':
                self.fatty_acid.set_isoform('iso')
            elif button.text == 'anteiso':
                self.fatty_acid.set_isoform('anteiso')

            if button.text in {'cis', 'trans', 'undefined'}:
                self.fatty_acid_mode = 'double bond'
                show_carbon_nr_buttons(self.screen, self.active_buttons, mode=self.fatty_acid_mode,
                                       fatty_acid=self.fatty_acid)

            show_buttons(FATTY_ACID_SUPER_OPTIONS_BUTTONS, self.screen, self.active_buttons)
            if self.fatty_acid_modified:
                hide_button(SET_ISOFORM_BUTTON, self.screen, self.active_buttons)

        elif type(button) == CButton:
            if self.fatty_acid_mode == 'double bond':
                self.fatty_acid.add_double_bond(button.nr, self.fatty_acid_bond_type)
                self.fatty_acid_bond_type = None
                self.fatty_acid_modified = True
            elif self.fatty_acid_mode == 'size':
                self.fatty_acid = FattyAcid(button.nr)
            elif self.fatty_acid_mode == 'amino':
                self.fatty_acid.add_amino_group(button.nr)
                self.fatty_acid_modified = True
            elif self.fatty_acid_mode == 'methyl':
                self.fatty_acid.add_methyl_group(button.nr)
                self.fatty_acid_modified = True
            elif self.fatty_acid_mode == 'hydroxyl':
                self.fatty_acid.add_hydroxyl_group(button.nr)
                self.fatty_acid_modified = True
            self.fatty_acid_mode = 'size'

            reset_buttons(self.screen, self.active_buttons)
            show_buttons(FATTY_ACID_SUPER_OPTIONS_BUTTONS, self.screen, self.active_buttons)
            if self.fatty_acid_modified:
                hide_button(SET_ISOFORM_BUTTON, self.screen, self.active_buttons)
            elif self.fatty_acid and self.fatty_acid.c_nr < 4:
                hide_button(SET_ISOFORM_BUTTON, self.screen, self.active_buttons)


        elif type(button) == AddGeneButton:
            button.do_action(self.screen, self.genes, self.text_box, self.active_buttons, mouse)
            self.text_box = None
        elif type(button) == CreateGeneButton:
            self.reset_selections()
            self.text_box = button.do_action(self.screen, self.active_buttons)
        elif type(button) == AddModuleButton:
            show_buttons([PKS_MODULE_BUTTON, NRPS_MODULE_BUTTON], self.screen, self.active_buttons)
        elif type(button) == PKSModuleButton or type(button) == NRPSModuleButton:
            button.do_action(self.screen, self.selected_gene, self.active_buttons, self.insertion_point, mouse)
            self.reset_selections()
            self.insertion_point = None
        elif type(button) == RemoveGeneButton:
            button.do_action(self.screen, self.genes, self.selected_gene, self.active_buttons, mouse)
        elif type(button) == AddDomainButton:
            reset_buttons(self.screen, self.active_buttons)
            show_domain_buttons(self.selected_module, self.screen, self.active_buttons)
        elif type(button) == RemoveModuleButton:
            button.do_action(self.screen, self.selected_module, self.active_buttons, mouse)
            self.reset_selections()
        elif type(button) == RemoveDomainButton:
            button.do_action(self.selected_domain, self.screen, self.active_buttons, mouse)
            self.reset_selections()
        elif type(button) == RenderProductsButton:

            try:

                render_products(self.genes, os.path.join(os.getcwd(), 'tmp_out_raichu'))
                product_dir = os.path.join(os.getcwd(), 'tmp_out_raichu')

                product_images = []

                for product_name in os.listdir(product_dir):
                    if product_name.endswith('.png'):
                        product_file = os.path.join(os.path.join(product_dir, product_name))
                        product_image = pygame.image.load(product_file)
                        product_images.append(product_image)

                product_image_height = RENDER_WINDOW_SIZE[1] / 2
                product_image_width = RENDER_WINDOW_SIZE[0] / 3

                for i, image in enumerate(product_images):
                    image_width = image.get_width()
                    image_height = image.get_height()

                    if image_width <= product_image_width and image_height > product_image_height:
                        ratio = product_image_height / image_height
                        image_width *= ratio
                        image_height = product_image_height
                    elif image_width > product_image_width and image_height <= product_image_height:
                        ratio = product_image_width / image_width
                        image_height *= ratio
                        image_width = product_image_width
                    elif image_width > product_image_width and image_height > product_image_height:
                        ratio_1 = product_image_width / image_width
                        ratio_2 = product_image_height / image_height
                        if ratio_1 <= ratio_2:
                            ratio = ratio_1
                        else:
                            ratio = ratio_2

                        image_height *= ratio
                        image_width *= ratio

                    render_window_size = (int(image_width), int(image_height))

                    self.product_images.append(pygame.transform.smoothscale(image, render_window_size))
                hide_buttons(ALL_BUTTONS, self.screen, self.active_buttons)
                show_button(SAVE_PRODUCTS_BUTTON, self.screen, self.active_buttons)

            except Exception as e:
                print(e)
                self.screen.blit(FAILED_TEXT_3, (0.42 * WIDTH, 0.52 * HEIGHT))
                self.screen.blit(FAILED_TEXT_2, (0.42 * WIDTH, 0.57 * HEIGHT))

            self.reset_selections()
        elif type(button) == RenderClusterButton:

            try:

                render_cluster(self.genes, os.path.join(os.getcwd(), 'tmp_out_raichu.png'))

                cluster_image = pygame.image.load(os.path.join(os.getcwd(), 'tmp_out_raichu.png'))

                image_width = cluster_image.get_width()
                image_height = cluster_image.get_height()

                if image_width <= RENDER_WINDOW_SIZE[0] and image_height > RENDER_WINDOW_SIZE[1]:
                    ratio = RENDER_WINDOW_SIZE[1] / image_height
                    image_width *= ratio
                    image_height = RENDER_WINDOW_SIZE[1]
                elif image_width > RENDER_WINDOW_SIZE[0] and image_height <= RENDER_WINDOW_SIZE[1]:
                    ratio = RENDER_WINDOW_SIZE[0] / image_width
                    image_height *= ratio
                    image_width = RENDER_WINDOW_SIZE[0]
                elif image_width > RENDER_WINDOW_SIZE[0] and image_height > RENDER_WINDOW_SIZE[1]:
                    ratio_1 = RENDER_WINDOW_SIZE[0] / image_width
                    ratio_2 = RENDER_WINDOW_SIZE[1] / image_height
                    if ratio_1 <= ratio_2:
                        ratio = ratio_1
                    else:
                        ratio = ratio_2

                    image_height *= ratio
                    image_width *= ratio

                render_window_size = (int(image_width), int(image_height))

                self.cluster_image = pygame.transform.smoothscale(cluster_image, render_window_size)
                hide_buttons(ALL_BUTTONS, self.screen, self.active_buttons)
                show_button(SAVE_CLUSTER_BUTTON, self.screen, self.active_buttons)
            except Exception as e:
                print(e)
                self.screen.blit(FAILED_TEXT_1, (0.42 * WIDTH, 0.52 * HEIGHT))
                self.screen.blit(FAILED_TEXT_2, (0.42 * WIDTH, 0.57 * HEIGHT))

            self.reset_selections()

        elif type(button) == SaveProductsButton or type(button) == SaveClusterButton:
            self.reset_selections()
            self.text_box = button.do_action(self.screen, self.active_buttons)
        elif type(button) == SaveToTxtButton:
            if self.text_box.text.endswith('.txt'):
                new_file = os.path.join(os.getcwd(), self.text_box.text)
            else:
                new_file = os.path.join(os.getcwd(), self.text_box.text + '.txt')

            if os.path.exists(new_file):
                for i, replace_text in enumerate(self.replace_text):
                    self.screen.blit(replace_text, (0.02 * WIDTH, 0.52 * HEIGHT + 0.05 * HEIGHT * i))
                show_buttons([YES_BUTTON, NO_BUTTON], self.screen, self.active_buttons)
            else:
                export_tabular(self.genes, new_file)
                self.random_click(mouse)

        elif type(button) == ExportTabularButton:
            self.reset_selections()
            self.text_box = button.do_action(self.screen, self.active_buttons)
        elif type(button) == SaveToPngButton:
            png_dir = os.path.join(os.getcwd(), 'tmp_out_raichu.png')
            if self.text_box.text.endswith('.png'):
                new_file = os.path.join(os.getcwd(), self.text_box.text)
            else:
                new_file = os.path.join(os.getcwd(), self.text_box.text + '.png')
            new_png_dir = os.path.join(os.getcwd(), new_file)
            if os.path.exists(new_png_dir):
                for i, replace_text in enumerate(self.replace_text):
                    self.screen.blit(replace_text, (0.02 * WIDTH, 0.52 * HEIGHT + 0.05 * HEIGHT * i))
                show_buttons([YES_BUTTON, NO_BUTTON], self.screen, self.active_buttons)
            else:
                os.replace(png_dir, new_png_dir)
                self.random_click(mouse)
        elif type(button) == SaveToFolderButton:
            folder = os.path.join(os.getcwd(), 'tmp_out_raichu')
            new_folder = os.path.join(os.getcwd(), self.text_box.text)
            if os.path.isdir(new_folder):
                for i, replace_text in enumerate(self.replace_text):
                    self.screen.blit(replace_text, (0.02 * WIDTH, 0.52 * HEIGHT + 0.05 * HEIGHT * i))
                show_buttons([YES_BUTTON, NO_BUTTON], self.screen, self.active_buttons)
            else:
                os.replace(folder, new_folder)
                self.random_click(mouse)

        elif type(button) == YesButton:
            if self.product_images:
                folder = os.path.join(os.getcwd(), 'tmp_out_raichu')
                new_folder = os.path.join(os.getcwd(), self.text_box.text)
                for file_name in os.listdir(new_folder):
                    file_path = os.path.join(new_folder, file_name)
                    os.remove(file_path)
                os.rmdir(new_folder)
                os.replace(folder, new_folder)
                self.random_click(mouse)
            elif self.cluster_image:
                png_dir = os.path.join(os.getcwd(), 'tmp_out_raichu.png')
                if self.text_box.text.endswith('.png'):
                    new_file = os.path.join(os.getcwd(), self.text_box.text)
                else:
                    new_file = os.path.join(os.getcwd(), self.text_box.text + '.png')

                os.replace(png_dir, new_file)
                self.random_click(mouse)
            else:
                if self.text_box.text.endswith('.txt'):
                    directory = os.path.join(os.getcwd(), self.text_box.text)
                else:
                    directory = os.path.join(os.getcwd(), self.text_box.text + '.txt')
                export_tabular(self.genes, directory)
                self.random_click(mouse)

        elif type(button) == NoButton:
            self.random_click(mouse)

        elif type(button) == SelectDomainTypeButton:
            reset_buttons(self.screen, self.active_buttons)
            show_buttons(KR_BUTTONS, self.screen, self.active_buttons)
        elif type(button) == SelectSubstrateButton:
            if self.selected_domain.type == 'A':
                reset_buttons(self.screen, self.active_buttons)
                show_buttons([STARTER_BUTTON, ELONGATION_BUTTON], self.screen, self.active_buttons)

            elif self.selected_domain.type == 'AT':
                reset_buttons(self.screen, self.active_buttons)
                show_buttons([STARTER_BUTTON, ELONGATION_BUTTON], self.screen, self.active_buttons)
        elif type(button) == KRSubtypeButton:
            button.do_action(self.selected_domain, mouse, self.screen, self.active_buttons)
            self.reset_selections()

        elif type(button) == SubstrateButton or type(button) == NRPSWildcardButton:
            self.selected_domain.substrate = button.substrate
            self.selected_domain.module.gene.erase()
            self.selected_domain.module.gene.draw(mouse)
            reset_buttons(self.screen, self.active_buttons)
            self.reset_selections()

    def do_click_action(self, mouse):

        selected_entity = self.detect_click_location(mouse)
        if selected_entity:
            for gene in self.genes:
                gene.draw(mouse)
            if type(selected_entity) == Button or issubclass(type(selected_entity), Button):
                self.do_button_action(selected_entity, mouse)
            elif type(selected_entity) == Gene:
                reset_buttons(self.screen, self.active_buttons)
                self.selected_gene = selected_entity
                self.insertion_point = InsertionPoint(selected_entity, mouse)
                selected_entity.insertion_point = self.insertion_point
                selected_entity.draw(mouse)
                show_buttons([ADD_MODULE_BUTTON, REMOVE_GENE_BUTTON], self.screen, self.active_buttons)
            elif type(selected_entity) == Module:
                self.selected_module = selected_entity
                reset_buttons(self.screen, self.active_buttons)
                show_buttons([ADD_DOMAIN_BUTTON, REMOVE_MODULE_BUTTON], self.screen,
                             self.active_buttons)
            elif type(selected_entity) == Domain:
                self.selected_domain = selected_entity
                reset_buttons(self.screen, self.active_buttons)
                show_buttons([REMOVE_DOMAIN_BUTTON], self.screen, self.active_buttons)

                if selected_entity.type == 'A' or selected_entity.type == 'AT':
                    show_button(SELECT_SUBSTRATE_BUTTON, self.screen, self.active_buttons)
                elif selected_entity.type == 'KR':
                    show_button(SELECT_DOMAIN_TYPE_BUTTON, self.screen, self.active_buttons)

        else:
            self.random_click(mouse)

    def random_click(self, mouse):
        self.screen.fill(WHITE)

        reset_buttons(self.screen, self.active_buttons)
        for gene in self.genes:
            gene.draw(mouse)

        for gene in self.genes:
            gene.selected = False
            gene.insertion_point = None
            for module in gene.modules:
                module.selected = False
                for domain in module.domains:
                    domain.selected = False

        self.selected_gene = None
        self.selected_module = None
        self.selected_domain = None
        self.selected_substrate_group = None
        self.cluster_image = False
        self.product_images = []
        self.text_box = None
        self.fatty_acid = None
        self.fatty_acid_mode = 'size'
        self.fatty_acid_bond_type = None
        self.fatty_acid_modified = False
        pygame.draw.rect(self.screen, WHITE, pygame.Rect(SIZE[0] - (FATTY_ACID_IMAGE_SIZE[0] + 20), SIZE[1] / 2 + 20,
                                                         FATTY_ACID_IMAGE_SIZE[0], FATTY_ACID_IMAGE_SIZE[1]))
        pygame.draw.rect(self.screen, WHITE, pygame.Rect(SIZE[0] - (FATTY_ACID_IMAGE_SIZE[0] + 20), SIZE[1] / 2 + 20,
                                                         FATTY_ACID_IMAGE_SIZE[0], FATTY_ACID_IMAGE_SIZE[1]), 2)
        raichu_dir = os.path.join(os.getcwd(), 'tmp_out_raichu')
        raichu_png_dir = os.path.join(os.getcwd(), 'tmp_out_raichu.png')
        if os.path.exists(raichu_png_dir):
            os.remove(raichu_png_dir)
        if os.path.exists(raichu_dir):
            for file_name in os.listdir(raichu_dir):
                file_path = os.path.join(raichu_dir, file_name)
                os.remove(file_path)
            os.rmdir(raichu_dir)










from pikachu.reactions.basic_reactions import condensation, hydrolysis
from pikachu.reactions.functional_groups import find_bonds, find_atoms
from raichu.class_domain import ATTRIBUTES
from raichu.central_chain_detection.label_central_chain import label_nrp_central_chain
from raichu.data.molecular_moieties import THIOESTERBOND, THIOESTER_CARBON, N_AMINO_ACID, B_N_AMINO_ACID
from raichu.reactions.general import label_rest_groups, initialise_atom_attributes, reset_nrp_annotations
from raichu.attach_to_domain import attach_to_domain_nrp


def nrps_elongation(amino_acid, chain_intermediate):
    """
    Returns the NRPS condensation product as a PIKAChU Structure object

    amino_acid: PIKAChU Structure object of the amino acid
    nrp_intermediate: PIKAChU Structure object of the NRP intermediate
    """

    # Initialize annotations in all atoms in amino acid
    initialise_atom_attributes(amino_acid)

    chain_intermediate, oh_bond = sulphur_to_hydroxyl(chain_intermediate)

    # Label rest groups
    label_rest_groups(amino_acid, chain_intermediate)

    # Ensure a central chain has been defined
    assert any(atom.annotations.in_central_chain for atom in chain_intermediate.graph)

    # Find amino group of amino acid or beta-amino acid
    n_atoms_aa = find_atoms(N_AMINO_ACID, amino_acid)
    if not n_atoms_aa:
        n_atoms_aa = find_atoms(B_N_AMINO_ACID, amino_acid)

    assert len(n_atoms_aa) == 1

    n_atom = n_atoms_aa[0]
    h_bond = None

    for bond in n_atom.bonds:
        for neighbour in bond.neighbours:
            if neighbour.type == 'H':
                h_bond = bond
                break

    assert h_bond

    # Label atoms in amino acid that end up in central peptide chain
    label_nrp_central_chain(amino_acid)

    # Reset chiral_c_ep and n_atom_nmeth AtomAnnotations attribute for all atoms in the intermediate;
    # This way, only the newly incorporated amino acid is labelled for tailoring
    reset_nrp_annotations(chain_intermediate)

    # Carry out condensation reaction using build-in PIKAChU function
    condensation_product = condensation(chain_intermediate, amino_acid, oh_bond, h_bond)[0]

    # Refresh condensation product
    condensation_product.refresh_structure(find_cycles=True)

    # Initialize annotations for atoms that don't have any yet
    initialise_atom_attributes(condensation_product)
    chain_intermediate = attach_to_domain_nrp(condensation_product)

    return chain_intermediate


def sulphur_to_hydroxyl(thioester_structure):
    """
    Identifies and removes the sulphur (and attached atoms/domains) in the
    thioester structure and replaces it with a hydroxyl group, creating a
    carboxylic acid group. Returns the product of this reaction, also as a
    PIKAChU Structure object.

    thioester_structure: PIKAChU Structure object of a thioester (R-C(S)=O)
    """
    # Find thioester bond in the input structure
    found_bonds_thioester = find_bonds(THIOESTERBOND, thioester_structure)
    found_carbon_thioester = find_atoms(THIOESTER_CARBON, thioester_structure)

    assert len(found_carbon_thioester) == 1
    assert len(found_bonds_thioester) == 1

    carbon_thioester = found_carbon_thioester[0]
    sh_bond = found_bonds_thioester[0]

    structures = hydrolysis(thioester_structure, sh_bond)
    combined = None
    for structure in structures:
        if carbon_thioester in structure.graph:
            combined = structure
            break

    assert combined

    oxygens = carbon_thioester.get_neighbours('O')

    oxygen_hydroxyl = None

    for oxygen in oxygens:
        if oxygen.get_bond(carbon_thioester).type == 'single' and oxygen.has_neighbour('H'):
            oxygen_hydroxyl = oxygen
            break

    assert oxygen_hydroxyl

    oh_bond = combined.bond_lookup[oxygen_hydroxyl][carbon_thioester]

    for atom in combined.graph:
        if not hasattr(atom.annotations, 'in_central_chain'):
            for attribute in ATTRIBUTES:
                atom.annotations.add_annotation(attribute, False)

    return combined, oh_bond




from pikachu.smiles.smiles import *
from pikachu.general import read_smiles
from raichu.central_chain_detection.label_central_chain import label_pk_central_chain


def find_central_chain(pks_nrps_attached):
    """Identifies the central chain atoms in the input structure from the
    in_central_chain Atom attribute, and returns these Atom objects as a list

    pks_nrps_attached: PIKAChU Structure object, input (hybrid) PK/NRP structure
    """
    # If the structure is a PK starter unit, find the central carbon chain

    if not any(atom.annotations.in_central_chain for atom in pks_nrps_attached.graph) and\
            len(pks_nrps_attached.find_substructures(read_smiles('C(=O)S'))) > 0:
        pks_nrps_attached = label_pk_central_chain(pks_nrps_attached)

    # Find atoms in the structure inside a cycle
    pks_nrps_attached.find_cycles()
    for atom in pks_nrps_attached.graph:
        if atom.in_ring(pks_nrps_attached):
            atom.inside_ring = True
        else:
            atom.inside_ring = False

    # Identify starting point central chain attached NRP/PK
    sulphur = None
    for atom in pks_nrps_attached.graph:
        if atom.type == 'S' and any(neighbour.annotations.domain_type
                                    for neighbour in atom.neighbours):
            sulphur = atom
        if not atom.annotations.in_central_chain:
            atom.annotations.in_central_chain = False
    assert sulphur

    central_chain = [sulphur]
    visited = [sulphur]
    atom_central_chain = sulphur
    end_atom = False

    # Identify complete central chain from in_central_chain Atom attributes
    while not end_atom:
        for neighbour in atom_central_chain.neighbours:
            if neighbour.annotations.in_central_chain and neighbour not in visited:
                central_chain.append(neighbour)
                visited.append(neighbour)
                atom_central_chain = neighbour
            elif not neighbour.annotations.in_central_chain:
                neighbours = []
                for next_atom in atom_central_chain.neighbours:
                    neighbours.append(next_atom)
                if all(atom in visited for atom in neighbours):
                    end_atom = True
                else:
                    visited.append(neighbour)

    return central_chain




from pikachu.reactions.functional_groups import GroupDefiner, find_atoms, find_bonds
from pikachu.general import read_smiles
from raichu.data.attributes import ATTRIBUTES
from raichu.reactions.general import initialise_atom_attributes
from raichu.attach_to_domain import attach_to_domain_pk
from raichu.data.molecular_moieties import AMINO_FATTY_ACID, AMINO_ACID_BACKBONE, N_AMINO_ACID, C1_AMINO_ACID, \
    C2_AMINO_ACID, BETA_AMINO_ACID_BACKBONE, B_N_AMINO_ACID, B_C1_AMINO_ACID, B_C2_AMINO_ACID, B_C3_AMINO_ACID, \
    B_LEAVING_BOND, ACID_C1

POLYKETIDE_S = GroupDefiner('Sulphur atom polyketide', 'SC(C)=O', 0)


def label_pk_central_chain(pk_starter_unit):
    """Finds the the atoms in the central chain of the polyketide starter unit,
    sets the in_central_chain Atom object attribute to True/False accordingly,
    after which the pk_starter_unit Structure object is returned.

    pk_starter_unit: PIKAChU Structure object of the PK starter unit
    """

    initialise_atom_attributes(pk_starter_unit)
    central_chain = []
    visited = []

    # Check if structure is a polyketide, define sulphur attached to domain
    locations_sulphur = find_atoms(POLYKETIDE_S, pk_starter_unit)
    assert len(locations_sulphur) == 1
    sulphur_polyketide = locations_sulphur[0]
    central_chain.append(sulphur_polyketide)

    # Iterate over atom neighbours, check if the atom belongs to the central
    # carbon chain (i.e., not a methyl/ethyl/methoxy sidechain) and add to list
    for neighbour in sulphur_polyketide.neighbours:
        if neighbour.type == 'C':
            chain_carbon = neighbour
            end_carbon = False
            central_chain.append(chain_carbon)
            visited.append(chain_carbon)

            while not end_carbon:
                for next_atom in chain_carbon.neighbours:
                    ethyl_branch = False
                    inside_cycle = False
                    methyl_group = False

                    # Build lists of neighbouring atom types
                    if next_atom.type == 'C' and next_atom not in visited:

                        # Keep track of carbon neighbours

                        # Confirm carbon doesn't belong to ethyl sidebranch
                        types = []
                        c_neighbours = next_atom.get_neighbours('C')
                        if len(c_neighbours) == 2 and next_atom.get_neighbour('O') is None:

                            for atom in c_neighbours:
                                for neighbour in atom.neighbours:
                                    types.append(neighbour.type)

                            if types.count('H') == 4 and types.count('C') == 4:
                                ethyl_branch = True

                            for atom in c_neighbours:
                                if atom not in visited and atom != next_atom:
                                    ethyl_branch = False

                            visited.append(chain_carbon)

                        # Carbon in a terminal carboxylic acid group is the
                        # final carbon in the central chain
                        if len(next_atom.get_neighbours('O')) == 2:
                            central_chain.append(next_atom)
                            end_carbon = True
                            visited.append(chain_carbon)

                        # Confirm carbon is not part of cycle
                        if len(c_neighbours) == 2:
                            if all(atom.in_ring(pk_starter_unit) for atom in c_neighbours):
                                print(15)
                                visited.append(next_atom)
                                inside_cycle = True
                                end_carbon = True

                        # Case if carbon is part of a benzene ring
                        if len(c_neighbours) == 3 and len(next_atom.neighbours) == 3:
                            # central_chain.append(next_atom) #added 10/04/2021
                            visited.append(next_atom)
                            inside_cycle = True
                            end_carbon = True

                        # Case where the PK starter ends in two methyl branches
                        if len(next_atom.get_neighbours('C')) == 3 and len(next_atom.get_neighbours('H')) == 1:
                            nr_methyl_brances = 0
                            methyl_carbon = None
                            for c_neighbour in c_neighbours:
                                c_neighbour_neighbour_types = []
                                for c_neighbour_neighbour \
                                        in c_neighbour.neighbours:
                                    c_neighbour_neighbour_types.append(
                                        c_neighbour_neighbour.type)
                                if c_neighbour_neighbour_types.count('H') == 3:
                                    methyl_carbon = c_neighbour
                                    nr_methyl_brances += 1

                            if nr_methyl_brances == 2:
                                assert methyl_carbon is not None
                                methyl_group = True
                                visited.append(next_atom)
                                central_chain.append(next_atom)
                                visited.append(methyl_carbon)
                                central_chain.append(methyl_carbon)
                                end_carbon = True

                        # Confirm carbon doesn't belong to methyl sidebranch
                        if len(next_atom.get_neighbours('H')) == 3 or \
                                (len(next_atom.get_neighbours('H')) == 2 and len(next_atom.get_neighbours('*')) == 1):
                            methyl_group = True

                            # Terminal methyl group is not a side branch:
                            count_c_neighbours_not_visited = 0
                            for neighbouring_atom in chain_carbon.neighbours:
                                if neighbouring_atom.type == 'C' and neighbouring_atom not in visited:
                                    count_c_neighbours_not_visited += 1
                            if count_c_neighbours_not_visited < 2:
                                methyl_group = False
                                visited.append(chain_carbon)
                                for neighbour in chain_carbon.neighbours:
                                    if neighbour.type == 'C' and neighbour not in visited:
                                        next_neighbours = []
                                        for next_neighbour in neighbour.neighbours:
                                            next_neighbours.append(
                                                next_neighbour.type)
                                        if next_neighbours.count('H') == 3 or (
                                                next_neighbours.count(
                                                        'H') == 2 and (
                                                next_neighbours.count(
                                                        '*')) == 1):
                                            central_chain.append(neighbour)
                                end_carbon = True

                            visited.append(chain_carbon)

                        # Identification of central chain carbon atoms through
                        # exclusion
                        if not ethyl_branch and not methyl_group and not \
                                end_carbon and not inside_cycle:
                            central_chain += [next_atom]
                            chain_carbon = next_atom
                            visited.append(chain_carbon)

    # Set in_central_chain Atom attribute
    for atom in pk_starter_unit.graph:
        if atom in central_chain:
            atom.annotations.in_central_chain = True
        else:
            atom.annotations.in_central_chain = False


def label_nrp_central_chain(peptide, module_type='elongation'):
    initialise_atom_attributes(peptide)

    as_normal = False
    if module_type == 'starter':
        if peptide.find_substructures(AMINO_FATTY_ACID):
            acid = peptide
            label_acid_central_chain(acid)
        else:
            as_normal = True

    if as_normal or module_type == 'elongation':

        if peptide.find_substructures(AMINO_ACID_BACKBONE):
            n_atoms_aa = find_atoms(N_AMINO_ACID, peptide)
            c1_atoms_aa = find_atoms(C1_AMINO_ACID, peptide)
            c2_atoms_aa = find_atoms(C2_AMINO_ACID, peptide)

            assert len(n_atoms_aa) == 1
            assert len(c1_atoms_aa) == 1
            assert len(c2_atoms_aa) == 1

            n_atoms_aa[0].annotations.in_central_chain = True
            n_atoms_aa[0].annotations.n_atom_nmeth = True
            c1_atoms_aa[0].annotations.in_central_chain = True
            c1_atoms_aa[0].annotations.chiral_c_ep = True
            c2_atoms_aa[0].annotations.in_central_chain = True
        elif peptide.find_substructures(BETA_AMINO_ACID_BACKBONE):
            n_atoms_aa = find_atoms(B_N_AMINO_ACID, peptide)
            c1_atoms_aa = find_atoms(B_C1_AMINO_ACID, peptide)
            c2_atoms_aa = find_atoms(B_C2_AMINO_ACID, peptide)
            c3_atoms_aa = find_atoms(B_C3_AMINO_ACID, peptide)

            assert len(n_atoms_aa) == 1
            assert len(c1_atoms_aa) == 1
            assert len(c2_atoms_aa) == 1
            assert len(c3_atoms_aa) == 1

            if n_atoms_aa[0].in_ring(peptide) and c1_atoms_aa[0].in_ring(peptide) and c2_atoms_aa[0].in_ring(peptide):
                label_acid_central_chain(peptide)

            else:

                n_atoms_aa[0].annotations.in_central_chain = True
                n_atoms_aa[0].annotations.n_atom_nmeth = True
                c1_atoms_aa[0].annotations.in_central_chain = True
                c1_atoms_aa[0].annotations.chiral_c_ep = True
                c2_atoms_aa[0].annotations.in_central_chain = True
                c3_atoms_aa[0].annotations.in_central_chain = True

                bonds = find_bonds(B_LEAVING_BOND, peptide)
                assert len(bonds) == 1
                bond = bonds[0]
                if bond.atom_1.type == 'O':
                    bond.atom_1.annotations.leaving_oh_o = True

                    bond.atom_2.annotations.leaving_oh_h = True

                elif bond.atom_2.type == 'O':
                    bond.atom_2.annotations.leaving_oh_o = True

                    bond.atom_1.annotations.leaving_oh_h = True

        else:
            # If NRPS starter unit is a (fatty) acid instead of an amino acid
            acid = peptide
            label_acid_central_chain(acid)


def label_acid_central_chain(acid):
    c1_atom = None
    c2_atom = None
    oh_o_atom = None
    oh_h_atom = None
    visited = []

    c1_atoms_aa = find_atoms(ACID_C1, acid)
    if c1_atoms_aa:
        c1_atom = c1_atoms_aa[0]

    assert c1_atom

    for neighbour in c1_atom.neighbours:
        neighbours = []

        if neighbour.type == 'C':
            for next_atom in neighbour.neighbours:
                neighbours.append(next_atom.type)
            if neighbours.count('O') == 2 and len(neighbours) == 3:
                c2_atom = neighbour

                # Find the leaving -OH group and annotate it

                for bond in c2_atom.bonds:
                    if bond.type == 'single' and (bond.has_neighbour('O')):
                        if bond.atom_1.type == 'O':
                            bond.atom_1.annotations.leaving_oh_o = True
                            oh_o_atom = bond.atom_1

                            bond.atom_2.annotations.leaving_oh_h = True
                            oh_h_atom = bond.atom_2

                        elif bond.atom_2.type == 'O':
                            bond.atom_2.annotations.leaving_oh_o = True
                            oh_o_atom = bond.atom_2

                            bond.atom_1.annotations.leaving_oh_h = True
                            oh_h_atom = bond.atom_1

    assert c2_atom
    assert oh_o_atom
    assert oh_h_atom

    c1_atom.annotations.in_central_chain = True
    c2_atom.annotations.in_central_chain = True
    c2_atom.annotations.c2_acid = True

    visited.append(c1_atom)
    visited.append(c2_atom)

    starting_point = c1_atom
    endpoint = False
    ring_members = 0
    if starting_point.in_ring(acid):
        ring_members += 1

    # Add atoms in chain with exactly 2 non-H neighbours to the central chain
    while not endpoint:
        for neighbour in starting_point.get_non_hydrogen_neighbours():
            neighbour_in_ring = neighbour.in_ring(acid)
            if neighbour_in_ring:
                ring_members += 1
            neighbour_found = False
            if len(neighbour.get_non_hydrogen_bonds()) == 2 and neighbour.type == 'C' and neighbour not in visited \
                    and not (ring_members > 2 and neighbour_in_ring):
                neighbour.annotations.in_central_chain = True
                visited.append(neighbour)
                starting_point = neighbour
                neighbour_found = True

            if not neighbour_found and neighbour.type == 'C' and neighbour not in visited and \
                    len(neighbour.get_non_hydrogen_bonds()) >= 2 and not (ring_members > 2 and neighbour_in_ring):
                terminal_count = 0
                for atom in neighbour.neighbours:
                    if atom not in visited and len(atom.get_non_hydrogen_bonds()) == 1:
                        terminal_count += 1

                if len(neighbour.get_non_hydrogen_bonds()) - terminal_count == 2:
                    neighbour.annotations.in_central_chain = True
                    visited.append(neighbour)
                    starting_point = neighbour
                    neighbour_found = True

                if neighbour_found:
                    break

            if not neighbour_found and len(neighbour.get_non_hydrogen_bonds()) == 2 and neighbour not in visited \
                    and not (ring_members > 2 and neighbour_in_ring):
                neighbour.annotations.in_central_chain = True
                visited.append(neighbour)
                starting_point = neighbour
                break

            neighbours = []
            for nb in starting_point.get_non_hydrogen_neighbours():
                if nb not in visited:
                    neighbours.append(nb)

            if any(atom.in_ring(acid) for atom in neighbours) and neighbour_in_ring:
                endpoint = True
                for atom in neighbours:
                    if len(atom.get_non_hydrogen_bonds()) == 1 and atom.get_non_hydrogen_bonds()[0].type == 'single':
                        atom.annotations.in_central_chain = True
                break

            elif not any(len(atom.get_non_hydrogen_bonds()) == 2 for atom in neighbours):

                endpoint = True

                for atom in neighbours:
                    terminal_count = 0
                    for atom_2 in atom.neighbours:
                        if atom_2 not in visited and len(atom_2.get_non_hydrogen_bonds()) == 1:
                            terminal_count += 1

                    if len(atom.get_non_hydrogen_bonds()) - terminal_count == 2:
                        endpoint = False
                        break

                if endpoint:

                    # Add the last atom in the chain to the central chain
                    for atom in neighbours:
                        if len(atom.get_non_hydrogen_bonds()) == 1 and atom.get_non_hydrogen_bonds()[0].type == 'single':
                            atom.annotations.in_central_chain = True
                    break


if __name__ == "__main__":
    starter_units_antismash = ['SC(=O)CC', 'SC(CC(O)=O)=O', 'SC(CC(O)=O)=O',
                               'SC(C(C(O)=O)CC)=O', 'SC(C(C(O)=O)OC)=O',
                               'SC(C*)=O',
                               'SC(C(C)CC)=O', 'SC(C1C(CCC1)C(=O)O)=O',
                               'SC(C)=O',
                               'SC(C1=CC=CC=C1)=O', 'SC(CC(C)C)=O',
                               'SC(C(C(=O)O)CC[Cl])=O']
    starter_unit_error = 'SC(C1=CC=CC=C1)=O'
    struct = read_smiles(starter_unit_error)
    struct.add_attributes(ATTRIBUTES, boolean=True)
    struct = attach_to_domain_pk(struct)
    label_pk_central_chain(struct)
    struct.print_graph()
    for atom in struct.graph:
        print(atom, atom.annotations.in_central_chain)


